package session_window;

use strict;

##############################################################################
# session_window package - create a scrollable window for viewing the 
# contents of a session from a tcpdump log file.
#
# exports create
#
# copyright (c) 1997 by Steve Romig and The Ohio State University
#
##############################################################################
# $Id: session_window.pm,v 1.4 1998/04/11 03:44:38 romig Exp $
##############################################################################
# $Log: session_window.pm,v $
# Revision 1.4  1998/04/11 03:44:38  romig
# changes to save file names, search paths for special files.
#
# Revision 1.3  1998/03/12 00:52:55  romig
# save file names can now be based on session within a log (host/port pair)
# as well as counter to make it unique.
#
# construct system directory pathnames (cache, report, summary...) from
# main::xyzzy since they can now be relative to the log file pathname
# to allow for separation between cases.
#
# Revision 1.1  1997/05/27 21:03:47  romig
# Initial revision
#

#
# special magic for some services
#
# this allows us to set specific defaults for certain services, to 
# make browsing as painless as possible.
#
%session_window::auto_magic = (
	       'irc' => {
		   'include' => '',
		   'exclude' => '^ison|^mode|^names|^who|^userhost|^#WAIT#| 302 | 324 ',
		   'rewrite' => 'irc_rewrite',
		   'width'   => 100,
		   'direction' => 'client',
	       },

	       'pop' => {
		   'include' => '^(From|Subject|To|Reply-To):',
		   'exclude' => '',
		   'rewrite' => '',
		   'width'   => 80,
		   'direction' => 'server',
	       },

	       'telnet' => {
		   'include' => '',
		   'exclude' => '',
		   'rewrite' => '',
		   'width'   => 80,
		   'direction' => 'client',
		   'replay' => 'replay_telnet',
	       },

	       'rlogin' => {
		   'include' => '',
		   'exclude' => '',
		   'rewrite' => '',
		   'width'   => 80,
		   'direction' => 'client',
		   'replay' => 'replay_telnet',
	       },

	       'x' => {
		   'include' => '',
		   'exclude' => '',
		   'rewrite' => '',
		   'width'   => 80,
		   'direction' => 'client',
		   'replay' => 'replay_x',
	       },

	       );

#
# given a service name, lookup info in the auto_magic array and return
# it, if anything.
#
sub get_magic {
    my($service) = @_;

    my $include = '';
    my $exclude = '';
    my $rewrite = '';
    my $width = 80;
    my $direction = 'client';
    my $replay = '';

    $include = $session_window::auto_magic{$service}{"include"} 
      if defined($session_window::auto_magic{$service}{"include"});

    $exclude = $session_window::auto_magic{$service}{"exclude"} 
      if defined($session_window::auto_magic{$service}{"exclude"});

    $rewrite = $session_window::auto_magic{$service}{"rewrite"} 
      if defined($session_window::auto_magic{$service}{"rewrite"});

    $width = $session_window::auto_magic{$service}{"width"} 
      if defined($session_window::auto_magic{$service}{"width"});

    $direction = $session_window::auto_magic{$service}{"direction"} 
      if defined($session_window::auto_magic{$service}{"direction"});

    $replay = $session_window::auto_magic{$service}{"replay"} 
      if defined($session_window::auto_magic{$service}{"replay"});

    return($include, $exclude, $rewrite, $width, $direction, $replay);
}

#
# rewrite text in the content window for an irc session to make names appear
# more readable.
# 
sub irc_rewrite {
    if ($_[0] =~ m/^([^:\s]+)\s+(.*)/s) {
	#
	# message client->server
	#
	# ignore a few things, or return as is
	#
	if ($1 eq 'PING' || $1 eq 'PONG' || $1 eq 'ISON' || 
	    $1 eq '#WAIT#') {
	    $_[0] = '';
	}
	
	return;
    }
	
    if ($_[0] =~ m/^:([^\s]+)\s+([0-9]+)/s) {
	#
	# numeric reply from server
	#
	$_[0] = '';

	return;
    }

    if ($_[0] =~ m/^:([^!]+)![^\s]*\s+([^\s]+)\s+(.*)$/) {
	#
	# message from server
	#

	if ($2 =~ m/pong/i) {
	    $_[0] = '';
	    return;
	}

	if ($2 eq 'PRIVMSG') {
	    $_[0] = "<$1> $3\n";
	} else {
	    $_[0] = "*** <$1> $2 $3\n";
	}

	return;
    }

    return;
}
    
#
# Run tcpdump on the indicated file to pull out client, server or 
# all traffic for a given connection
#
# $box is the box to fill, $file is the file to read the data from, 
# $what is 'all', 'client' or 'server'.  the rest of the args should
# be obvious.  src and dst are relative to the initial tcp syn packet
# from the summary, so client is src and server is dst.
#

sub redraw_session {
    my($box, $file, $what, $src_ip, $src_port, $dst_ip, $dst_port, 
	  $include, $exclude, $rewrite, $use_rewrite) = @_;
    my $selector = '';
    my $e_include = '';
    my $e_exclude = '';

    if ($what eq "all") {
	$selector = "host $src_ip and port $src_port and host $dst_ip and port $dst_port";
    } elsif ($what eq "client") {
	$selector = "src host $src_ip and src port $src_port and dst host $dst_ip and dst port $dst_port";
    } else {
	$selector = "dst host $src_ip and dst port $src_port and src host $dst_ip and src port $dst_port";
    }

    if ($include ne '') {
	$e_include = "$main::egrep '$include' |";
    }

    if ($exclude ne '') {
	$e_exclude = "$main::egrep -v '$exclude' |";
    }

    my $cache_file = &cache::get_cache_file_name($file, $src_ip, $src_port, 
						 $dst_ip, $dst_port, 0);

    my $command = "$main::tcpdump_extract -r $cache_file '$selector' | $main::cleanup_text | $e_include $e_exclude";

    print $command, "\n"
	if $main::debug;

    open(TCP, $command) || die "couldn't open tcpdump\n";

    bless $box, "Tk::Text";

    # delete everything in the listbox and replace with what we read from
    # tcpdump
    $box->configure(-state => "normal");
    $box->delete('1.0', 'end');

    $_ = '';			# silence the critic
    while (<TCP>) {
	no strict 'refs';
	s/\r\n/\n/;
	&$rewrite($_) if $use_rewrite && $rewrite ne '';
	next if ($_ eq "\n" || $_ eq '');
	$box->insert('end', $_);
    }

    close(TCP);
    $box->configure(-state => "disabled");
}

sub replay_telnet {
    my($file, $src_ip, $src_port, $dst_ip, $dst_port) = @_;

    my $server_port = &main::min($src_port, $dst_port);
    my $cache_file = &cache::get_cache_file_name($file, $src_ip, $src_port, 
						 $dst_ip, $dst_port, 0);
    my $tmp_file = "$cache_file.slowcat";

    system("$main::tcpdump_extract -r $cache_file src port $server_port  | $main::cleanup_literal > $tmp_file");
    &exec::fork_and_exec("$main::xterm -sb -sk -sl 500 -bg white -e $main::slowcat $tmp_file");
    #
    # fixme: $tmp_file should be cleaned up, but its in the cache dir
    # so that's probably ok as it will be deleted when that's cleaned
    #
}

sub replay_x {
    my($file, $src_ip, $src_port, $dst_ip, $dst_port) = @_;

    my $cache_file = &cache::get_cache_file_name($file, $src_ip, 
						 $src_port, $dst_ip, 
						 $dst_port, 0);

    system("$main::tcpdump_extract -r $cache_file src port $dst_port | $main::cleanup_literal > $cache_file.xreplay.server");
    system("$main::tcpdump_extract -r $cache_file src port $src_port | $main::cleanup_literal > $cache_file.xreplay.client");

    if (! open(DEPTH, "$main::rpx $cache_file.xreplay.client $cache_file.xreplay.server |")) {
	warn "can't examine initial X server state\n";
	return;
    }

    my ($x, $y, $depth) = (0, 0, 0);
    my $display;

    while (<DEPTH>) {
	if ($_ =~ m/([0-9]+) x ([0-9]+) x ([0-9]+)/) {
	    $x = $1;
	    $y = $2;
	    $depth = $3;
	    last;
	}
    }
    close(DEPTH);

    if ($depth != 0 && defined($ENV{"DISPLAY$depth"})) {
	$display = $ENV{"DISPLAY$depth"};
    } elsif (defined($ENV{'DISPLAY'})) {
	$display = $ENV{'DISPLAY'};
    } else {
	$display = 'localhost:0';
    }
    print "replay depth is $depth, using $display\n";

    &exec::fork_and_exec("$main::xterm -sb -si -sl 1000 -e $main::rpx -replay -display=$display $cache_file.xreplay.client $cache_file.xreplay.server");
}

#
# do the real work of saving session contents.
#
sub do_save {
    my ($file, $src_ip, $src_port, $dst_ip, $dst_port, $savefile, 
	$direction, $decoding, $window) = @_;
    my $selector = '';
    my $cleaner = '';

    if (-f $savefile) {
	warn "$savefile exists!  will not overwrite\n";
	return;
    }

    my $cache_file = &cache::get_cache_file_name($file, $src_ip, $src_port, 
						 $dst_ip, $dst_port, 0);

    if ($direction eq 'client') {
	$selector = "src port $src_port";
    } elsif ($direction eq 'server') {
	$selector = "src port $dst_port";
    }

    if ($decoding eq 'ascii') {
	$cleaner = $main::cleanup_text;
    } else {
	$cleaner = $main::cleanup_literal;
    }

    my $command = "$main::tcpdump_extract -r $cache_file $selector | $cleaner > $savefile";

    print $command, "\n"
	if $main::debug;

    system $command;

    $window->destroy;
}

#
# pop up a dialogue box in preparation for saving session contents
#
sub save {
    my($file, $direction, $src_ip, $src_port, $dst_ip, $dst_port) = @_;
    
    my $decoding = 'literal';
    my $extension;

    if ($main::use_long_save_name) {
	$extension = ".$src_ip\_$src_port\_$dst_ip\_$dst_port";
    } else {
	$extension = '';
    }

    my $counter = 0;

    my $base_savefile = &main::xyzzy($main::save_dir, 
				     $file, 
				     "$extension.$counter");
    my $savefile = $base_savefile;

    while (-f $savefile) {
	$savefile = $base_savefile . '.' . $counter++;
    }

    my $save_window = $log_window::main->Toplevel;

    $save_window->title("Saving $file");

    my $stupid_top_button_frame = $save_window->Frame;

    #
    # 2 choices for decoding: save filtered printable ascii, or 
    # literal contents
    #

    my $select_decoding_frame = $stupid_top_button_frame->Frame;

    my $ascii = 
	$select_decoding_frame->Radiobutton(-text => 'ascii',
					    -variable => \$decoding,
					    -value => 'ascii',
					    -anchor => 'w');
    my $literal = 
	$select_decoding_frame->Radiobutton(-text => 'literal',
					    -variable => \$decoding,
					    -value => 'literal',
					    -anchor => 'w');

    $ascii->pack(-side => 'top', -fill => 'x');
    $literal->pack(-side => 'top', -fill => 'x');

    $select_decoding_frame->pack(-side => 'left');

    #
    # 3 choices for direction: save client->server traffic, server->client 
    # or both
    #

    my $select_direction_frame = $stupid_top_button_frame->Frame;

    my $client = 
	$select_direction_frame->Radiobutton(-text => 'client',
					     -variable => \$direction,
					     -value => 'client',
					     -anchor => 'w');
    my $server = 
	$select_direction_frame->Radiobutton(-text => 'server',
					     -variable => \$direction,
					     -value => 'server',
					     -anchor => 'w');
    my $both = 
	$select_direction_frame->Radiobutton(-text => 'both',
					     -variable => \$direction,
					     -value => 'both',
					     -anchor => 'w');

    $client->pack(-side => 'top', -fill => 'x');
    $server->pack(-side => 'top', -fill => 'x');
    $both->pack(-side => 'top', -fill => 'x');

    $select_direction_frame->pack(-side => 'left');
    
    #
    # text entry for the file name (should use a browser)
    #
    my $file_frame = $save_window->Frame;

    my $file_label = $file_frame->Label(-text => 'File name');
    my $file_entry = $file_frame->Entry(-width => 40,
					-textvariable => \$savefile);

    $file_label->pack(-side => 'left');
    $file_entry->pack(-side => 'left');

    #
    # doit and cancel buttons
    #
    my $button_frame = $save_window->Frame;

    my $doit_button = 
	$button_frame->Button(-text => 'Save it',
			      -command => sub{&do_save($file, 
						       $src_ip,
						       $src_port,
						       $dst_ip,
						       $dst_port,
						       $savefile, 
						       $direction, 
						       $decoding,
						       $save_window)});

    my $cancel_button = 
	$button_frame->Button(-text => 'Cancel',
			      -command => ['destroy', $save_window]);

    $doit_button->pack(-side => 'left');
    $cancel_button->pack(-side => 'left');

    #
    # pack 'em up
    #
    $stupid_top_button_frame->pack(-side => 'top');
    $file_frame->pack(-side => 'top');
    $button_frame->pack(-side => 'top');
}

sub find_session {
    my($file, $ip_a, $port_a, $ip_b, $port_b, 
       $target_proto, $target_ip, $target_port) = @_;
    my $line;
    my($packets, $time, $cand_ip_a, $cand_port_a, $cand_ip_b, $cand_port_b, 
       $proto, $tags);

    #
    # cough, sputter.  read through the summary in search 
    # of a session that has a matching ip/port to get the 
    # ip/port of the other end.
    #
    foreach $line (&summary::get($file, "$main::egrep '$target_proto' |", '')) {
	($packets, $time, $cand_ip_a, $cand_port_a, $cand_ip_b, $cand_port_b, 
	   $proto, $tags) = &summary::parse($line);

	if (($target_ip eq $cand_ip_a && $target_port eq $cand_port_a) ||
	    ($target_ip eq $cand_ip_b && $target_port eq $cand_port_b)) {
	    return($proto, $cand_ip_a, $cand_port_a, $cand_ip_b, $cand_port_b);
	}
    }

    return;
}

#
# jump to a new session.  icky stuff here.  
#
sub jump {
    my($file, $frame, $src_ip, $src_port, $dst_ip, $dst_port) = @_;
    my $line;
    my $ip;
    my $new_ip;
    my $new_port;
    my $target_ip;
    my $target_port;
    my($new_proto, $new_src_ip, $new_src_port, $new_dst_ip, $new_dst_port);

    chop($line = $frame->get('sel.first', 'sel.last'));

    # if the line looks like an ftp port command, try to jump to
    # the indicated session in this log...
    #
    if ($line =~ m/PORT (\d+,\d+,\d+,\d+),(\d+),(\d+)/) {
	$new_ip = $1;
	$new_port = $2 * 256 + $3;

	if ($new_ip eq $dst_ip) {
	    $new_src_ip = $src_ip;
	    $new_src_port = 20;
	    $new_dst_ip = $dst_ip;
	    $new_dst_port = $new_port;
	} else {
	    $new_src_ip = $dst_ip;
	    $new_src_port = 20;
	    $new_dst_ip = $src_ip;
	    $new_dst_port = $new_port;
	}	    

	&create($file, 'tcp', $new_src_ip, $new_src_port, 
		$new_dst_ip, $new_dst_port);
	return;
    }

    # if the line looks like an ftp pasv response, try to jump to
    # the indicated session in this log...
    #
    if ($line =~ m/^227 .*\((\d+,\d+,\d+,\d+),(\d+),(\d+)\)/) {
	$target_ip = $1;
	$target_ip =~ tr/,/./;
	$target_port = $2 * 256 + $3;

	($new_proto, $new_src_ip, $new_src_port, 
	             $new_dst_ip, $new_dst_port) =
	    &find_session($file, $src_ip, $src_port, $dst_ip, $dst_port,
			  'tcp', $target_ip, $target_port);

	# fixme: check for failure
	# fixme: pass $proto
	&create($file, $new_proto, $new_src_ip, $new_src_port, 
		$new_dst_ip, $new_dst_port);

	return;
    }

    # if the line looks like irc dcc commands, try to jump to it...
    #
    if ($line =~ m/DCC SEND .* (\d+) (\d+) \d+/ || 
	$line =~ m/DCC CHAT [^ \t]+ (\d+) (\d+)/) {
	$ip = $1+0;
	$target_port = $2;

	$target_ip = join('.', ($ip >> 24) & 0xff, ($ip >> 16) & 0xff,
			   ($ip >> 8) & 0xff, $ip & 0xff);

	($new_proto, $new_src_ip, $new_src_port, $new_dst_ip, $new_dst_port) =
	    &find_session($file, $src_ip, $src_port, $dst_ip, $dst_port,
			  'tcp', $target_ip, $target_port);

	if (defined($new_proto) && $new_proto ne "") {
	    &create($file, $new_proto, $new_src_ip, $new_src_port, 
		    $new_dst_ip, $new_dst_port);
	    
	    return;
	}
    }

    print "sorry, couldn't find the associated session\n";
}		


#
# create a new window that contains the traffic data for browsing
#

sub create {
    my($file, $proto, $src_ip, $src_port, $dst_ip, $dst_port) = @_;
    
    my $which_traffic = 'client';
    my $include = '';
    my $exclude = '';
    my $rewrite = '';
    my $width = 80;
    my $replay;
    my $replay_button;
    
    my $use_rewrite = 1;

    # try to guess the type for this session - note that we can't
    # simply pass the value from the summary in, since in some cases
    # we don't necessarily have a type to pass in.
    #
    # though on second thought, in those cases we can infer the type from
    # the context.  e.g., calling create from jump in an ftp session 
    # implies that the next session is also ftp.
    #
    my($to_server, $type, $group) = &services::get_tags($proto, $src_ip, 
							$src_port, $dst_ip, 
							$dst_port);

    if ($type ne '') {
	($include, $exclude, $rewrite, $width, $which_traffic, $replay) = 
	    &get_magic($type);
    }

    my $window = $log_window::main->Toplevel;

    $window->title("Session: $src_ip $src_port <-> $dst_ip $dst_port");

    # defined now in case we need it for references?
    #
    my $text = $window->Text(-relief => 'sunken',
			     -background => $main::background_color,
			     -foreground => $main::foreground_color,
			     -width => $width,
			     -height => 40,
			     -setgrid => 'yes');

    # first row of buttons
    #
    # +-----------------------------
    # | dismiss save redraw replay...
    # |
    #
    my $buttons = $window->Frame;
    $buttons->pack;

      # dismiss button
      #
      my $dismiss_button = 
	  $buttons->Button(-text => 'Dismiss',
			   -command => ['destroy', $window]);
      $dismiss_button->pack(-side => 'left');

      # save button
      #
      my $save_button = 
	  $buttons->Button(-text => 'Save',
			   -command => sub{&save($file, 
						 $which_traffic,
						 $src_ip, 
						 $src_port,
						 $dst_ip,
						 $dst_port)});
      $save_button->pack(-side => 'left');

      # redraw button
      #
      my $redraw_button = 
	  $buttons->Button(-text => 'Redraw',
			   -command => sub{&redraw_session($text, 
							 $file, 
							 $which_traffic, 
							 $src_ip, 
							 $src_port, 
							 $dst_ip, 
							 $dst_port,
							 $include,
							 $exclude,
							 $rewrite, 
							 $use_rewrite)});
      $redraw_button->pack(-side => 'left');

      # replay button (if applicable)
      
      if (defined($replay) && $replay ne '') {
	  no strict 'refs';
	  $replay_button = 
	      $buttons->Button(-text => 'Replay',
			       -command => sub{&$replay($file,
							$src_ip,
							$src_port,
							$dst_ip,
							$dst_port)});
	  $replay_button->pack(-side => 'left');
      }

      # "show all" radio button
      #
      my $show_all = 
	  $buttons->Radiobutton(-text => 'all',
				-variable => \$which_traffic,
				-value => 'all',
				-anchor => 'w',
				-command => sub{&redraw_session($text, 
							      $file, 
							      $which_traffic, 
							      $src_ip, 
							      $src_port, 
							      $dst_ip, 
							      $dst_port,
							      $include,
							      $exclude,
							      $rewrite,
							      $use_rewrite)});
      $show_all->pack(-side => 'left');

      # "show client" radio button
      #
      my $show_client = 
	  $buttons->Radiobutton(-text => 'client only',
				-variable => \$which_traffic,
				-value => 'client',
				-anchor => 'w',
				-command => sub{&redraw_session($text, 
							      $file, 
							      $which_traffic, 
							      $src_ip, 
							      $src_port, 
							      $dst_ip, 
							      $dst_port,
							      $include,
							      $exclude,
							      $rewrite,
							      $use_rewrite)});
      $show_client->pack(-side => 'left');

      # "show server" radio button
      #
      my $show_server = 
	  $buttons->Radiobutton(-text => 'server only',
				-variable => \$which_traffic,
				-value => 'server',
				-anchor => 'w',
				-command => sub{&redraw_session($text, 
							      $file, 
							      $which_traffic, 
							      $src_ip, 
							      $src_port, 
							      $dst_ip, 
							      $dst_port,
							      $include,
							      $exclude,
							      $rewrite,
							      $use_rewrite)});
      $show_server->pack(-side => 'left');

      # "user rewrite" checkbox
      #
      my $use_rewrite_button = 
	  $buttons->Checkbutton(-text => 'use rewrite',
				-variable => \$use_rewrite,
				-command => sub{&redraw_session($text, 
							      $file, 
							      $which_traffic, 
							      $src_ip, 
							      $src_port, 
							      $dst_ip, 
							      $dst_port,
							      $include,
							      $exclude,
							      $rewrite,
							      $use_rewrite)});
      $use_rewrite_button->pack(-side => "left");

    # create a frame for the include grep entry boxes
    #
    # +-----------------------------
    # | dismiss save redraw replay...
    # | include...           exclude...
    #
    my $inc_exc_frame = $window->Frame;

      # include box stuff...
      #
      my $include_frame = $inc_exc_frame->Frame;

      my $label_include = $include_frame->Label(-text => 'Include');
      my $entry_include = $include_frame->Entry(-width => 40,
                                                -background => 
						    $main::background_color,
                                                -foreground => 
						    $main::foreground_color,
                                                -textvariable => \$include);

      $label_include->pack(-side => 'left');
      $entry_include->pack(-side => 'left');

      $include_frame->pack(-side => 'left');

      # exclude box stuff...
      #
      my $exclude_frame = $inc_exc_frame->Frame;

      my $label_exclude = $exclude_frame->Label(-text => 'Exclude');
      my $entry_exclude = $exclude_frame->Entry(-width => 40,
                                                -background => 
						    $main::background_color,
                                                -foreground => 
						    $main::foreground_color,
					        -textvariable => \$exclude);

      $label_exclude->pack(-side => 'left');
      $entry_exclude->pack(-side => 'left');

      $exclude_frame->pack(-side => 'right');

    $inc_exc_frame->pack;

    # create the scroll bar
    #
    # +-----------------------------
    # | dismiss save redraw replay...
    # | include...           exclude...
    # + +-----------------------+ S
    # | | summary text here     | C
    # | |                       | R
    # | |                       | O
    # | |                       | L
    # | +-----------------------+ L
    #
    $text->pack(-side => 'left',
		-fill => 'both',
		-expand => 'yes');

    my $scroll = $window->Scrollbar(-command => ['yview', $text]);

    $text->configure(-yscrollcommand => ['set', $scroll]);

    $scroll->pack(-side => 'right',
		  -fill => 'y');

    # set bindings...using triple-click for now, since we are lazy
    # and haven't figured out how to get the line of text any other way.
    #
    # triple button jumps to an inferred session in this log.  see 
    # comments on &jump for details
    #
    $text->bind("<Triple-Button-1>", 
		sub{jump($file, $text, $src_ip, $src_port, 
			 $dst_ip, $dst_port)});

    # fill the text box with stuff
    #
    &redraw_session($text, $file, $which_traffic, $src_ip, $src_port,
		  $dst_ip, $dst_port, $include, $exclude, $rewrite, 
		  $use_rewrite); 
}

1;
