/* 
 * read packets from tcpdump format log and display the contents
 *
 * usage: cat file | tcp-print [-f] [-r]
 */

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <pcap.h>

#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>

#define ETHER_HDR_SIZE sizeof(struct ether_header);

#ifdef HEX
#define PHEX(p,l)	print_hex(p, l)
#else
#define PHEX(p,l)
#endif

#define RBSIZE	8192		/* size of initial buffer, increment size */

int debug = 0;
int filter = 0;
int reassemble = 0;
int doneone = 0;
int option_i = 0;
u_short cvt_buffer[2048];

/* 
 * tcp reassembly stuff
 */
unsigned long seq;		/*  */
unsigned long isn = 0;		/*  */
unsigned long last = 0;		/*  */
unsigned char *buffer;		/* tcp reassembly buffer */
unsigned long buffersize = 0;	/* size of reassembly buffer */

extern char *optarg;

usage() 

{
  fprintf(stderr, "usage: tcp-print [-dfri] [file] \n");
  exit(1);
}

u_char *reorder(const u_char *buffer, int bytes)

{
  int i, words;
  const u_short *src = (u_short *)buffer;
  
  words = (bytes / 2) + ((bytes % 2) ? 1 : 0);

  for (i = 0; i < words; i++) {
    cvt_buffer[i] = ntohs(*(src + i));
  }

  return((u_char *) cvt_buffer);
}

print_hex(const u_char *data, int bytes)

{
  int i = 0, j;
  const u_short *rover = (u_short *)data;

  putchar('\n');
  
  while (i * 2 < bytes) {
    for (j = 0; j < 8; j++) {
      if (i * 2 < bytes)
	printf("%04x ", *(rover + i++));
    }

    putchar('\n');
  }
}

print_data(const u_char *data, int bytes_left) 

{
  const u_char *rover;

  if (debug) fputs("data:", stdout);

  if (option_i) {
    fputs("|data ", stdout);
  }

  for (rover = data; rover < data + bytes_left; rover++) {
    if (filter) {
      /*
       * in filter mode, print all printable ascii characters, 
       * plus \n and \r.  this does funny things to escape sequences 
       * and the like.  life is rough. 
       */
      switch (*rover) {
      case 8:			/* ^H backspace */
	fputs("^H", stdout);
	break;

      case 9:			/* ^I tab */
	putchar(*rover);
	break;

      case '\n':		/* ^J linefeed */
	if (option_i) {
	  fputs("\\n", stdout);
	} else {
	  putchar(*rover);
	}
	break;

      case '\r':		/* ^M return */
	fputs("\\r", stdout);
	break;

      case 127:			/* ^? delete */
	fputs("^?", stdout);
	break;

      default:
	if ((*rover > 31) && (*rover < 127))
	  putchar(*rover);
	break;
      }
    } else {
      /* else print the literal data */
      putchar(*rover);
    }
  }

  if (option_i) {
    putchar('\n');
  }
}

print_reassemble (struct tcphdr *tcp_hdr, const u_char *data, int bytes_left)

{
  int seq;
  unsigned long offset;	/* offset from start of stream to this data */
  unsigned char *save_buffer;

  seq = ntohl(tcp_hdr->th_seq);

  /*
   * fixme:
   * first packet better be the real one to get isn set correctly!
   */
  if (isn == 0) {
    isn = seq + 1;		/* set isn if it isn't already */
  }

  offset = seq - isn;

  if (debug) 
    printf("  %lu %lu offset %lu length %lu last %lu\n", 
	   seq, isn, offset, bytes_left, last);

  if (bytes_left > 0) {
    /* grow the buffer if we need to */
    if (offset + bytes_left > buffersize) {
      buffersize += RBSIZE;
      save_buffer = buffer;
      buffer = (char *) realloc(buffer, buffersize);
      if (buffer == NULL) {
	fprintf(stderr, 
		"realloc of reassembly buffer failed at %d bytes\n", 
		buffersize);
	
	/* try to write what we have */
	if (write(fileno(stdout), save_buffer, last) < 0) {
	  perror("write failed");
	}

	exit(1);
      }
    }

    /* 
     * copy the data portion of the packet to the buffer.
     * the idea here is that we just write stuff 
     * willy-nilly to the buffer, hopefully filling in
     * the whole thing, without worrying about the order
     * in which we received the data.  duplicate packets 
     * should take care of themselves, and corrupted data
     * should be replaced when we see the re-xmitted
     * packet later.
     */
    bcopy(data, (buffer+offset), bytes_left);

    if (offset + bytes_left > last) last = offset + bytes_left;
    /* 
     * fixme: keep track of the segments that are filled
     * in so we can detect missing segments
     */
  }
}

print_tcp(const u_char *p, int bytes_left)

{
  struct tcphdr *tcp_hdr;
  const u_char *data;

  if (bytes_left < sizeof(struct tcphdr))
    return;

  if (debug) fputs("tcp...", stdout);

  tcp_hdr = (struct tcphdr *) p;
  data = p + sizeof(struct tcphdr);
  bytes_left -= sizeof(struct tcphdr);

  if (option_i) {
    printf("|tcp %u %u %2x %lu ",
	   ntohs(tcp_hdr->th_sport),
	   ntohs(tcp_hdr->th_dport),
	   tcp_hdr->th_flags,
	   ntohl(tcp_hdr->th_seq));
  }

  if (reassemble)
    print_reassemble(tcp_hdr, data, bytes_left);
  else
    print_data(data, bytes_left);
}

print_udp(const u_char *p, int bytes_left)

{
  const u_char *data;
  struct udphdr *udp_hdr;

  if (bytes_left < sizeof(struct udphdr))
    return;

  if (debug) fputs("udp...", stdout);

  udp_hdr = (struct udphdr *)p;

  data = p + sizeof(struct udphdr);
  bytes_left -= sizeof(struct udphdr);

  if (option_i) {
    printf("|udp %u %u ",
	   ntohs(udp_hdr->uh_sport),
	   ntohs(udp_hdr->uh_dport));
  }

  print_data(data, bytes_left);
}

print_icmp(const u_char *p, int bytes_left)

{
  const u_char *data;
  struct icmp *icmp_hdr;
  char *what;

  if (bytes_left < ICMP_MINLEN)
    return;

  if (debug) fputs("icmp...", stdout);

  icmp_hdr = (struct icmp *)p;

  /* fixme: should calculate lengths to skip to data also */
  switch (icmp_hdr->icmp_type) {
  case ICMP_ECHOREPLY:
    what = "echo request";
    break;

  case ICMP_UNREACH:
    switch(icmp_hdr->icmp_code) {
    case ICMP_UNREACH_NET:
      what = "unreach net";
      break;

    case ICMP_UNREACH_HOST:
      what = "unreach host";
      break;

    case ICMP_UNREACH_PROTOCOL:
      what = "unreach proto";
      break;

    case ICMP_UNREACH_PORT:
      what = "unreach port";
      break;

    case ICMP_UNREACH_NEEDFRAG:
      what = "unreach needfrag";
      break;

    case ICMP_UNREACH_SRCFAIL:
      what = "unreach srcfail";
      break;

    case ICMP_UNREACH_NET_UNKNOWN:
      what = "unreach netunknown";
      break;

    case ICMP_UNREACH_HOST_UNKNOWN:
      what = "unreach hostunknown";
      break;

    case ICMP_UNREACH_ISOLATED:
      what = "unreach isolated";
      break;

    case ICMP_UNREACH_NET_PROHIB:
      what = "unreach netprohib";
      break;

    case ICMP_UNREACH_HOST_PROHIB:
      what = "unreach hostprohib";
      break;

    case ICMP_UNREACH_TOSNET:
      what = "unreach tosnet";
      break;

    case ICMP_UNREACH_TOSHOST:
      what = "unreach toshost";
      break;

    case ICMP_UNREACH_FILTER_PROHIB:
      what = "unreach filterprohib";
      break;

    case ICMP_UNREACH_HOST_PRECEDENCE:
      what = "unreach hostprec";
      break;

    case ICMP_UNREACH_PRECEDENCE_CUTOFF:
      what = "unreach cutoffprec";
      break;

    default:
      what = "unreach unknown";
    }
    break;

  case ICMP_SOURCEQUENCH:
    what = "srcquench";
    break;

  case ICMP_REDIRECT:
    what = "redirect";
    break;

  case ICMP_ECHO:
    what = "echo reply";
    break;

  case ICMP_ROUTERADVERT:
    what = "route adv";
    break;

  case ICMP_ROUTERSOLICIT:
    what = "route solicit";
    break;

  case ICMP_TIMXCEED:
    what = "time exceeded";
    break;

  case ICMP_PARAMPROB:
    what = "param problem";
    break;

  case ICMP_TSTAMP:
    what = "time stamp";
    break;

  case ICMP_TSTAMPREPLY:
    what = "time stamp reply";
    break;

  case ICMP_IREQ:
    what = "ireq";
    break;

  case ICMP_IREQREPLY:
    what = "ireq reply";
    break;

  case ICMP_MASKREQ:
    what = "mask request";
    break;

  case ICMP_MASKREPLY:
    what = "mask reply";
    break;

  default:
    what = "unknown";
    break;
  }

  if (option_i) {
    printf("|icmp %s ",
	   what);
  }

  /* fixme: should skip icmp header, etc.  ick. */

  print_data(data, bytes_left);
}

print_ip(const u_char *p, int bytes_left)

{
  struct ip *ip_hdr;
  int hlen;
  const u_char *data;
  char src[20], dst[20];

  if (bytes_left < 20) /* min ip hdr size */
    return;

  if (debug) fputs("ip...", stdout);
  PHEX(p, bytes_left);

  ip_hdr = (struct ip *) p;
  hlen = ip_hdr->ip_hl * 4;

  data = p + hlen;
  bytes_left -= hlen;

  if (option_i) {
    strncpy(src, inet_ntoa(ip_hdr->ip_src), 20);
    strncpy(dst, inet_ntoa(ip_hdr->ip_dst), 20);

    printf("|ip %s %s %d %d %02x ",
	   src,
	   dst,
	   ip_hdr->ip_hl,
	   ntohs(ip_hdr->ip_off),
	   ip_hdr->ip_p);
  }

  switch (ip_hdr->ip_p) {
  case IPPROTO_TCP:
    print_tcp(data, bytes_left);
    break;

  case IPPROTO_UDP:
    print_udp(data, bytes_left);
    break;

  case IPPROTO_ICMP:
    print_icmp(data, bytes_left);
    break;

  default:
    printf("@ip layer: doesn't handle type %x yet\n", ip_hdr->ip_p);
    break;
  }
}

print_ether(const u_char *p, int bytes_left) 

{
  struct ether_header *e_hdr;
  const u_char *data;

  if (bytes_left < sizeof(struct ether_header))
    return;

  if (debug) fputs("ether...", stdout);
  PHEX(p, bytes_left);

  e_hdr = (struct ether_header *) p;
  bytes_left -= sizeof(struct ether_header);
  data = p + sizeof(struct ether_header);

  switch (ntohs(e_hdr->ether_type)) {
  case ETHERTYPE_IP:
    print_ip(data, bytes_left);
    break;

  default:
    printf("@ether layer: doesn't handle type %x yet\n", 
	    e_hdr->ether_type);
    break;
  }
}

print_time(long sec, long usec) 

{
  struct tm *lt;
  
  lt = localtime(&sec);
  printf("%02d%02d.%02d:%02d:%02d.%06d ",
	 lt->tm_mon+1,
	 lt->tm_mday,
	 lt->tm_hour,
	 lt->tm_min,
	 lt->tm_sec,
	 usec);
}

/* 
 * called to act on each packet from the packet log.
 * skips the mac layer header (assumed to be ethernet for now) and calls
 * the per-proto print routine for the next layer.
 */
blargh(char *user, const struct pcap_pkthdr *h, const u_char *p)

{
  const u_char *b;

#ifdef REORDER
  if (doneone++) return;

  PHEX(p, h->caplen);
  b = reorder(p, h->caplen);
  PHEX(b, h->caplen);
#else
  b = p;
#endif /* REORDER */

  /* fixme: if option_i print timestamp */
  if (option_i) {
    print_time(h->ts.tv_sec, h->ts.tv_usec);
  }

  /* fixme: how to determine type of header? */
  print_ether(b, h->caplen);

  if (debug) printf("\n");
}

main(argc, argv)
  int argc;
  char ** argv;

{
  char		*file_name;
  pcap_t	*pd;
  char		errbuf[8192];
  char		c;

  while ((c = getopt(argc, argv, "dfri")) != -1) {
    switch (c) {
    case 'd':
      debug = 1;
      break;

    case 'f':
      filter = 1;
      break;
      
    case 'r':
      reassemble = 1;
      break;

    case 'i':
      option_i = 1;
      break;

    default:
      usage();
      break;
    }
  }

  file_name = "-";

  if (reassemble && option_i) {
    fprintf(stderr, "sorry, can't use -i and -r both\n");
    exit(1);
  }

  if (reassemble) {
    buffer = (char *) malloc(RBSIZE); /* initial buffer is 8k */
    buffersize = RBSIZE;	/* ditto */

    if (buffer == NULL) {
      fprintf(stderr, "malloc for reassembly buffer failed\n");
      exit(1);
    }
  }

  /*
   * open the log file, dealing with byte ordering crap
   */
  pd = pcap_open_offline(file_name, errbuf);
  if (pd == NULL) {
    fprintf(stderr, "error: %s\n", errbuf);
    exit(1);
  }

  /* read the log, "print" each packet back to the net */
  pcap_loop(pd, -1, (pcap_handler) blargh, NULL);

  /* close the log, bpf device */
  pcap_close(pd);

  if (reassemble) {
    if (write(fileno(stdout), buffer, last) < 0) {
      perror("write failed");
    }
  }

}

